<?xml version="1.0" encoding="UTF-8"?>
<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
  version="2.0">
  
  <!-- An XSLT 2.0 stylesheet that generates, for an XML document,
       warnings noting the appearance of unknown characters
       ('known characters' is a set that can be extended)
    
    This could be done in XSLT 1.0, but 2.0 is faster and easier
    
    Creates plain-text output. A fancy version for production use
      might create formatted (HTML or PDF) output
  
  -->
  
  <xsl:output method="text"/>

  <xsl:template name="header">
    <xsl:value-of select="document-uri(/)"/>
    <xsl:for-each select="/paper/front/author">
      <xsl:text>&#xA;</xsl:text>
      <xsl:value-of select="normalize-space(string-join((fname|surname), ' '))"/>
      <xsl:if test="position() != last()">
        <xsl:choose>
          <xsl:when test="position() + 1 = last()"> and </xsl:when>
          <xsl:otherwise>, </xsl:otherwise>
        </xsl:choose>
      </xsl:if>
    </xsl:for-each>
    <xsl:text>&#xA;</xsl:text>
    <xsl:apply-templates select="/paper/front/title" mode="plain"/>
    <xsl:text>&#xA;--- </xsl:text>
  </xsl:template>
  
  <xsl:template match="/" name="extreme-chars">
    <xsl:call-template name="header"/>
    <xsl:apply-templates select="//text()" mode="character-warnings"/>
    <xsl:text>&#xA;=====&#xA;</xsl:text>
  </xsl:template>

  <xsl:variable name="okay">
    abcdefghijklmnopqrstuvwxyz
    ABCDEFGHIJKLMNOPQRSTUVWXYZ
    1234567890
    `~!@#$%^&amp;*()-_=+
    []{}\|;:&quot;&apos;,./&lt;&gt;?
    
    &#xA0;
  </xsl:variable>

  <xsl:variable name="Unicode-lookup" select="document('Unicode-codepoint-lookup.xml')"/>
  
  <xsl:key name="char-by-codepoint" match="char" use="cp"/>
  <!-- indexed retrieval of <char> element (in lookup table)
    by its hex codepoint (a string) -->
  
  <xsl:template match="text()" mode="character-warnings">
    <xsl:variable name="unknowns" select="translate(.,$okay,'')"/>
    <xsl:if test="$unknowns">
      <xsl:text>&#xA;Warning: character(s) out of bounds: </xsl:text>
      <xsl:text>&#xA;  in path </xsl:text>
      <xsl:apply-templates select="." mode="make-path"/>
      <xsl:text>&#xA;</xsl:text>
      <xsl:for-each select="1 to string-length($unknowns)">
        <xsl:variable name="char" select="substring($unknowns,.,1)"/>
        <xsl:variable name="donechars" select="substring($unknowns,1,(. - 1))"/>
        <xsl:if test="not(contains($donechars, $char))">
          <xsl:text> '</xsl:text>
          <xsl:value-of select="$char"/>
          <xsl:text>' U+</xsl:text>
          <xsl:variable name="codepoint">
            <xsl:variable name="cp-value">
            <xsl:call-template name="num2hex">
              <xsl:with-param name="num" select="string-to-codepoints($char)"/>
            </xsl:call-template>
            </xsl:variable>
            <xsl:value-of
              select="substring('0000',(1+ string-length($cp-value)))"/>
            <xsl:value-of select="$cp-value"/>
          </xsl:variable>
          <xsl:value-of select="$codepoint"/>
          <xsl:text> (</xsl:text>
          <xsl:value-of select="key('char-by-codepoint',$codepoint,$Unicode-lookup)/name"/>
          <xsl:text>)&#xA;</xsl:text>
        </xsl:if>
      </xsl:for-each>
    </xsl:if>
  </xsl:template>  

  <xsl:template name="num2hex">
    <!-- this should be implemented as a function, but it was
         easier to adapt a template found on line:
         thanks to Aristotle Pagaltzis at plasmasturm.org
         (http://plasmasturm.org/log/386/) -->
    <xsl:param name="num"/>
    <xsl:variable name="nybble" select="$num mod 16"/>
    <xsl:variable name="head" select="floor( $num div 16 )"/>
    <xsl:variable name="rest">
      <xsl:if test="$head">
        <xsl:call-template name="num2hex">
          <xsl:with-param name="num" select="$head"/>
        </xsl:call-template>
      </xsl:if>
    </xsl:variable>
    <xsl:value-of select="$rest"/>
    <xsl:choose>
      <xsl:when test="$nybble > 9">
        <xsl:value-of select="translate( substring( string($nybble), 2 ), '012345', 'ABCDEF' )"/>
      </xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="$nybble"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>

  <xsl:include href="xpath-support.xsl"/>

</xsl:stylesheet>
